# Writing pytest tests for the DataStorageBlock class

import pytest

from .data_storage_code_block import DataStorageCodeBlock


def test_create_and_retrieve_item():
    storage = DataStorageCodeBlock()
    storage.create("test_key", "test_value")
    assert storage.retrieve("test_key") == "test_value"


def test_update_item():
    storage = DataStorageCodeBlock()
    storage.create("test_key", "test_value")
    storage.update("test_key", "new_test_value")
    assert storage.retrieve("test_key") == "new_test_value"


def test_delete_item():
    storage = DataStorageCodeBlock()
    storage.create("test_key", "test_value")
    storage.delete("test_key")
    assert storage.retrieve("test_key") is None


def test_create_existing_key():
    storage = DataStorageCodeBlock()
    storage.create("test_key", "test_value")
    with pytest.raises(ValueError):
        storage.create("test_key", "another_value")


def test_update_nonexistent_key():
    storage = DataStorageCodeBlock()
    with pytest.raises(ValueError):
        storage.update("nonexistent_key", "value")


def test_list_all_items():
    storage = DataStorageCodeBlock()
    storage.create("key1", "value1")
    storage.create("key2", "value2")
    all_items = storage.list_all()
    assert all_items == {"key1": "value1", "key2": "value2"}



import os
import pytest
from typing import Optional
from unittest import mock
from urllib.parse import urlparse


TEST_AUTH = 'testauth123'
TEST_CONFIG = {
    "SOME_KEY": "SOME_VALUE",
}

class MockResponse:
    def __init__(self, text: Optional[str], status_code: int):
        self.text = text
        self.status_code = status_code

    def __repr__(self):
        return f"<{__class__.__name__} status_code={self.status_code} text={self.text}>"

    @property
    def ok(self):
        return 200 <= self.status_code < 300

def mock_get(url, **kwargs):
    if kwargs.get('headers', {}).get('Authorization', '') != "Bearer " + TEST_AUTH:
        return MockResponse(None, 401)
    p = urlparse(url)
    if not p.path.startswith("/secrets/v1/"):
        return MockResponse(None, 404)
    key = p.path[len("/secrets/v1/"):]
    if key in TEST_CONFIG:
        return MockResponse(TEST_CONFIG[key], 200)
    else:
        return MockResponse(None, 404)


@pytest.fixture(autouse=True)
def mock_svcprovider():
    os.environ['BISMUTH_AUTH'] = TEST_AUTH
    with mock.patch.multiple('requests', get=mock_get):
        yield


def test_get_existant_key():
    from .configuration_code_block import ConfigurationCodeBlock
    config = ConfigurationCodeBlock()
    assert config.get('SOME_KEY') == "SOME_VALUE"


def test_get_nonexistant_key():
    from .configuration_code_block import ConfigurationCodeBlock
    with pytest.raises(AttributeError):
        _ = ConfigurationCodeBlock().get('NONEXISTANT')



__version__ = "0.1.11"

from .codeblocks.api_code_block import *
from .codeblocks.api_code_block_with_storage import *
from .codeblocks.auth_code_block import *
from .codeblocks.base_code_block import *
from .codeblocks.configuration_code_block import *
from .codeblocks.data_storage_code_block import *
from .codeblocks.function_code_block import *
from .codeblocks.persistent_data_storage_code_block import *



import os
import pytest
from flask import Flask
from unittest.mock import MagicMock
from .api_code_block_with_storage import (
    APICodeBlockWithStorage,
)
from .data_storage_code_block import DataStorageCodeBlock


@pytest.fixture
def app():
    app = Flask(__name__)
    return app


@pytest.fixture
def api_block_with_storage(app):
    os.environ['BISMUTH_AUTH'] = "TEST_AUTH"
    api_block = APICodeBlockWithStorage()
    api_block.app.testing = True
    return api_block


def test_add_route_with_storage(api_block_with_storage):
    mock_data_storage = MagicMock(spec=DataStorageCodeBlock)
    mock_data_storage.retrieve.return_value = {"reponse": "OK"}
    api_block_with_storage.add_route_with_storage("/mock", "GET", mock_data_storage)

    with api_block_with_storage.app.test_client() as client:
        client.get("/mock?key=test")
        mock_data_storage.retrieve.assert_called_with("test")



import json
import os
import pytest
from typing import Optional, Dict
from unittest import mock
from urllib.parse import urlparse

from .persistent_data_storage_code_block import PersistentDataStorageCodeBlock


TEST_AUTH = 'testauth123'


class MockResponse:
    def __init__(self, data: Optional[str | bytes], status_code: int):
        self.data = data
        self.status_code = status_code

    def __repr__(self):
        return f"<{__class__.__name__} status_code={self.status_code} json={repr(self.json)}>"

    def json(self):
        if self.data is None:
            return None
        return json.loads(self.data)

    @property
    def ok(self):
        return 200 <= self.status_code < 300

def mock_get(storage, url, **kwargs):
    if kwargs.get('headers', {}).get('Authorization', '') != TEST_AUTH:
        return MockResponse(None, 401)
    p = urlparse(url)
    if p.path == "/blob/v1/":
        # svcprovider JSONifies the entire KV store here, and Rust JSON serializes byte arrays as [int].
        return MockResponse(json.dumps({k: list(v) for k,v in storage.items()}), 200)
    elif p.path.startswith("/blob/v1/"):
        key = p.path[len("/blob/v1/"):]
        if key in storage:
            return MockResponse(storage[key], 200)
        else:
            return MockResponse(None, 404)

def mock_post(storage, url, **kwargs):
    if 'Authorization' not in kwargs.get('headers', {}):
        return MockResponse(None, 401)
    p = urlparse(url)
    key = p.path[len("/blob/v1/"):]
    if key in storage:
        return MockResponse(None, 409)
    else:
        storage[key] = bytes(kwargs['data'], 'utf-8')
        return MockResponse(None, 200)

def mock_put(storage, url, **kwargs):
    if 'Authorization' not in kwargs.get('headers', {}):
        return MockResponse(None, 401)
    p = urlparse(url)
    key = p.path[len("/blob/v1/"):]
    if key in storage:
        storage[key] = bytes(kwargs['data'], 'utf-8')
        return MockResponse(None, 200)
    else:
        return MockResponse(None, 404)

def mock_delete(storage, url, **kwargs):
    if 'Authorization' not in kwargs.get('headers', {}):
        return MockResponse(None, 401)
    p = urlparse(url)
    key = p.path[len("/blob/v1/"):]
    if key in storage:
        del storage[key]
        return MockResponse(None, 200)
    else:
        return MockResponse(None, 404)


@pytest.fixture(autouse=True)
def mock_svcprovider():
    os.environ['BISMUTH_AUTH'] = TEST_AUTH
    storage: Dict[str, bytes] = {}
    with mock.patch.multiple('requests',
                             get=lambda url, **kwargs: mock_get(storage, url, **kwargs),
                             post=lambda url, **kwargs: mock_post(storage, url, **kwargs),
                             put=lambda url, **kwargs: mock_put(storage, url, **kwargs),
                             delete=lambda url, **kwargs: mock_delete(storage, url, **kwargs)):
        yield


def test_create_and_retrieve_item():
    storage = PersistentDataStorageCodeBlock()
    storage.create("test_key", "test_value")
    assert storage.retrieve("test_key") == "test_value"


def test_update_item():
    storage = PersistentDataStorageCodeBlock()
    storage.create("test_key", "test_value")
    storage.update("test_key", "new_test_value")
    assert storage.retrieve("test_key") == "new_test_value"


def test_delete_item():
    storage = PersistentDataStorageCodeBlock()
    storage.create("test_key", "test_value")
    storage.delete("test_key")
    assert storage.retrieve("test_key") is None


def test_create_existing_key():
    storage = PersistentDataStorageCodeBlock()
    storage.create("test_key", "test_value")
    with pytest.raises(ValueError):
        storage.create("test_key", "another_value")


def test_update_nonexistent_key():
    storage = PersistentDataStorageCodeBlock()
    with pytest.raises(ValueError):
        storage.update("nonexistent_key", "value")


def test_list_all_items():
    storage = PersistentDataStorageCodeBlock()
    storage.create("key1", "value1")
    storage.create("key2", "value2")
    all_items = storage.list_all()
    assert all_items == {"key1": "value1", "key2": "value2"}



import os
import requests
from flask import request
from http import HTTPStatus
from typing import Optional
from urllib.parse import urljoin

from .base_code_block import BaseCodeBlock


class ConfigurationCodeBlock(BaseCodeBlock):
    """
    The ConfigurationCodeBlock provides access to configuration parameters set during creation.
    Any configured key is available through the get() method, returning the string value or raising an exception if no
    such configuration key exists.
    """

    def __init__(self, api_url="http://169.254.169.254:9000/secrets/v1/"):
        if 'BISMUTH_AUTH' not in os.environ:
            print("Missing BISMUTH_AUTH token in environment. Configuration will not be available.")
            return
        self.auth = {"Authorization": "Bearer " + os.environ['BISMUTH_AUTH']}
        self.api_url = api_url

    def _headers(self):
        hdrs = self.auth.copy()
        try:
            for tracehdr in ["traceparent", "tracestate"]:
                if tracehdr in request.headers:
                    hdrs[tracehdr] = request.headers[tracehdr]
        except RuntimeError:
            pass
        return hdrs

    def get(self, key) -> Optional[str]:
        if 'BISMUTH_AUTH' not in os.environ:
            raise Exception("Missing BISMUTH_AUTH token in environment.")
        resp = requests.get(urljoin(self.api_url, key), headers=self._headers())
        if resp.status_code == HTTPStatus.NOT_FOUND:
            raise AttributeError(f"Configuration key {key} not set")
        elif not resp.ok:
            raise Exception(f"Server error {resp}")
        return resp.text



from .base_code_block import BaseCodeBlock
from typing import Callable, Any


class FunctionCodeBlock(BaseCodeBlock):
    def __init__(
        self, func: Callable[[Any], Any], network_enabled: bool = False, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.network_enabled = network_enabled
        self.func = func

    def exec(self, **kwargs):
        return self.func(**kwargs)



from functools import wraps
from flask import request, jsonify
from .base_code_block import BaseCodeBlock


class AuthCodeBlock(BaseCodeBlock):
    def __init__(self, auth_service=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if auth_service is None:
            self.auth_service = MockAuthService()
        else:
            self.auth_service = auth_service

    def token_required(self, f):
        @wraps(f)
        def decorated(*args, **kwargs):
            token = None

            if "Authorization" in request.headers:
                auth_header = request.headers["Authorization"]
                if auth_header.startswith("Bearer "):
                    token = auth_header.split(" ")[1]

            if not token:
                return jsonify({"message": "Token is missing!"}), 401

            if not self.auth_service.validate_token(token):
                return jsonify({"message": "Token is invalid!"}), 401

            return f(*args, **kwargs)

        return decorated


# Example of a mocked authentication service
class MockAuthService:
    @staticmethod
    def validate_token(token):
        # Replace this with real validation logic
        return token == "valid_token"



from flask import Flask, request
from flask_restx import Api, Resource
from .auth_code_block import AuthCodeBlock
from .base_code_block import BaseCodeBlock
from .configuration_code_block import ConfigurationCodeBlock
from .function_code_block import FunctionCodeBlock


class APICodeBlock(BaseCodeBlock):
    def __init__(
        self,
        title="API",
        version="1.0",
        description="A simple API",
        config: ConfigurationCodeBlock = None,
        auth_code_block: AuthCodeBlock = None,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.app = Flask(__name__)
        @self.app.route("/healthz")
        def healthz():
            return "ok"

        if config is None:
            config = ConfigurationCodeBlock()

        self.config = config
        # Hack: set prefix so that this doesn't register over the root route
        self.api = Api(self.app, version=version, title=title, description=description, doc="/doc", prefix="/tmp")
        # And clear it
        self.api.prefix = ""
        self.auth_code_block = auth_code_block

    def _auth_handler(
        self,
        method: str,
        handlers: dict[str, FunctionCodeBlock],
        require_auth: list[str],
        **kwargs
    ):
        cb = handlers[method].exec if isinstance(handlers[method], FunctionCodeBlock) else handlers[method]

        if (
            method.upper() in map(lambda x: x.upper(), require_auth)
            and self.auth_code_block is not None
        ):
            return self.auth_code_block.token_required(cb)(**kwargs)
        else:
            return cb(**kwargs)

    def add_route(
        self,
        route: str,
        handlers: dict[str, FunctionCodeBlock],
        require_auth: list[str] = ["POST", "DELETE", "PUT"],
    ):
        handlers = {k.upper(): v for k, v in handlers.items()}
        auth_handler = self._auth_handler

        class DynamicResource(Resource):
            def get(self):
                kwargs = request.args.to_dict()
                if "GET" in handlers:
                    return auth_handler("GET", handlers, require_auth, **kwargs)
                else:
                    self.api.abort(404)

            def post(self):
                kwargs = request.json or {}
                if "POST" in handlers:
                    return auth_handler("POST", handlers, require_auth, **kwargs)
                else:
                    self.api.abort(404)

            def put(self):
                kwargs = request.json or {}
                if "PUT" in handlers:
                    return auth_handler("PUT", handlers, require_auth, **kwargs)
                else:
                    self.api.abort(404)

            def delete(self):
                kwargs = request.args.to_dict()

                if "DELETE" in handlers:
                    return auth_handler("DELETE", handlers, require_auth, **kwargs)
                else:
                    self.api.abort(404)

        self.api.add_resource(DynamicResource, route)

    def __call__(self, *args):
        return self.app(*args)

    def run(self, host="0.0.0.0", port=5000, debug=False):
        self.app.run(host=host, port=port, debug=debug)



from typing import Callable
from flask import Flask, request
from flask_restx import Api, Resource
from .api_code_block import APICodeBlock
from .data_storage_code_block import DataStorageCodeBlock
from .auth_code_block import AuthCodeBlock
from .function_code_block import FunctionCodeBlock


class APICodeBlockWithStorage(APICodeBlock):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_stores = {}  # Dictionary to hold DataStorageBlock instances

    def _auth_handler_for_storage(
        self, method, func: FunctionCodeBlock | Callable, require_auth, **kwargs
    ):
        cb = func.exec if isinstance(func, FunctionCodeBlock) else func

        if (
            method.upper() in map(lambda x: x.upper(), require_auth)
            and self.auth_code_block is not None
        ):
            return self.auth_code_block.token_required(func.exec)(**kwargs)
        else:
            return func.exec(**kwargs)

    def add_route_with_storage(
        self,
        route,
        methods,
        data_storage_block: DataStorageCodeBlock,
        require_auth=["POST", "DELETE", "PUT"],
        auth_code_block: AuthCodeBlock = None,
    ):
        self.data_stores[route] = data_storage_block
        methods = methods if isinstance(methods, list) else [methods]
        methods = [method.upper() for method in methods]

        if auth_code_block is None:
            self.auth_code_block = AuthCodeBlock()
        else:
            self.auth_code_block = auth_code_block

        auth_handler_for_storage = self._auth_handler_for_storage

        class DynamicResource(Resource):
            def get(self):
                if "GET" in methods:
                    def run():
                        key = request.args.get("key")
                        return (
                            data_storage_block.retrieve(key)
                            if key
                            else data_storage_block.list_all()
                        )

                    func = FunctionCodeBlock(run)

                    return auth_handler_for_storage("GET", func, require_auth)
                else:
                    self.api.abort(405)

            def post(self):
                if "POST" in methods:
                    def run():
                        data = request.json
                        key, value = data.get("key"), data.get("value")
                        if key and value:
                            data_storage_block.create(key, value)
                            return {"message": "Item created"}, 201
                        else:
                            return {"message": "Key and value required"}, 400

                    func = FunctionCodeBlock(run)
                    return auth_handler_for_storage("POST", func, require_auth)
                else:
                    self.api.abort(405)

            def put(self):
                if "PUT" in methods:
                    def run():
                        data = request.json
                        key, value = data.get("key"), data.get("value")
                        if key and value:
                            data_storage_block.update(key, value)
                            return {"message": "Item updated"}
                        else:
                            return {"message": "Key and value required"}, 400

                    return auth_handler_for_storage("PUT", run, require_auth)
                else:
                    self.api.abort(405)

            def delete(self):
                self._auth_handler()
                if "DELETE" in methods:
                    def run():
                        key = request.args.get("key")
                        if key:
                            data_storage_block.delete(key)
                            return {"message": "Item deleted"}
                        else:
                            return {"message": "Key required"}, 400

                    func = FunctionCodeBlock(run)
                    return auth_handler_for_storage("DELETE", func, require_auth)

        self.api.add_resource(DynamicResource, route)



from .base_code_block import BaseCodeBlock


class DataStorageCodeBlock(BaseCodeBlock):
    def __init__(self):
        """Initialize the datastore with an empty dictionary."""
        self.store = {}

    def create(self, key, value):
        """Create a new item in the datastore."""
        if key in self.store:
            raise ValueError("Key already exists.")
        self.store[key] = value

    def retrieve(self, key):
        """Retrieve an item from the datastore."""
        return self.store.get(key, None)

    def update(self, key, value):
        """Update an existing item in the datastore."""
        if key not in self.store:
            raise ValueError("Key does not exist.")
        self.store[key] = value

    def delete(self, key):
        """Delete an item from the datastore."""
        if key in self.store:
            del self.store[key]

    def list_all(self):
        """List all items in the datastore."""
        return self.store



import os
import pytest
from unittest.mock import MagicMock
from .api_code_block import APICodeBlock
from .function_code_block import FunctionCodeBlock


# Fixture to setup the APICodeBlock with mocked routes
@pytest.fixture
def api_block():
    os.environ['BISMUTH_AUTH'] = "TEST_AUTH"
    api_block = APICodeBlock()
    api_block.app.testing = True
    return api_block


def test_add_route(api_block):
    mock_block = MagicMock(spec=FunctionCodeBlock)
    mock_block.exec.return_value = {"message": "mock response"}
    api_block.add_route("/mock", {"get": mock_block})

    with api_block.app.test_client() as client:
        response = client.get("/mock")
        assert response.status_code == 200
        assert response.json == {"message": "mock response"}
        mock_block.exec.assert_called_once()


def test_add_route_bare_func(api_block):
    def func():
        return {"message": "mock response"}
    api_block.add_route("/mock", {"get": func})

    with api_block.app.test_client() as client:
        response = client.get("/mock")
        assert response.status_code == 200
        assert response.json == {"message": "mock response"}


def test_add_root_route(api_block):
    mock_block = MagicMock(spec=FunctionCodeBlock)
    mock_block.exec.return_value = {"message": "mock response"}
    api_block.add_route("/", {"get": mock_block})

    with api_block.app.test_client() as client:
        response = client.get("/")
        assert response.status_code == 200
        assert response.json == {"message": "mock response"}
        mock_block.exec.assert_called_once()



import json
import os
import requests
from flask import request
from http import HTTPStatus
from typing import Optional, Dict, Any
from urllib.parse import urljoin

from .data_storage_code_block import DataStorageCodeBlock


class PersistentDataStorageCodeBlock(DataStorageCodeBlock):
    """
    The PersistentDataStorageCodeBlock uses Bismuth's blob storage service to persist JSON-encodable objects.
    """
    def __init__(self, api_url="http://169.254.169.254:9000/blob/v1/"):
        if 'BISMUTH_AUTH' not in os.environ:
            raise Exception("Missing BISMUTH_AUTH token in environment.")
        self.auth = {"Authorization": os.environ['BISMUTH_AUTH']}
        self.api_url = api_url

    def _headers(self):
        hdrs = self.auth.copy()
        try:
            for tracehdr in ["traceparent", "tracestate"]:
                if tracehdr in request.headers:
                    hdrs[tracehdr] = request.headers[tracehdr]
        except RuntimeError:
            pass
        return hdrs

    def create(self, key, value) -> None:
        """Create a new item in the datastore."""
        resp = requests.post(urljoin(self.api_url, key), data=json.dumps(value), headers=self._headers())
        if resp.status_code == HTTPStatus.CONFLICT:
            raise ValueError("Key already exists.")
        elif not resp.ok:
            raise Exception(f"Server error {resp}")

    def retrieve(self, key) -> Optional[Any]:
        """Retrieve an item from the datastore."""
        resp = requests.get(urljoin(self.api_url, key), headers=self._headers())
        if resp.status_code == HTTPStatus.NOT_FOUND:
            return None
        elif not resp.ok:
            raise Exception(f"Server error {resp}")
        return resp.json()

    def update(self, key, value) -> None:
        """Update an existing item in the datastore."""
        resp = requests.put(urljoin(self.api_url, key), data=json.dumps(value), headers=self._headers())
        if resp.status_code == HTTPStatus.NOT_FOUND:
            raise ValueError("Key does not exist.")
        elif not resp.ok:
            raise Exception(f"Server error {resp}")

    def delete(self, key) -> None:
        """Delete an item from the datastore."""
        resp = requests.delete(urljoin(self.api_url, key), headers=self._headers())
        if resp.status_code == HTTPStatus.NOT_FOUND:
            raise ValueError("Key does not exist.")
        elif not resp.ok:
            raise Exception(f"Server error {resp}")

    def list_all(self) -> Dict[str, Any]:
        """List all items in the datastore."""
        resp = requests.get(self.api_url, headers=self._headers())
        if not resp.ok:
            raise Exception(f"Server error {resp}")
        return dict((k, json.loads(bytes(v))) for k, v in resp.json().items())
